import abc
import collections
from functools import partial
import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax import jax_utils
from flax.training.common_utils import stack_forest
import jax
from jax import (
    lax,
    random,
    value_and_grad,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from jax.tree_util import tree_map
from tqdm import trange
from tqdm.std import tqdm as TqdmDecorator

from fortuna.data.loader import (
    DataLoader,
    TargetsLoader,
)
from fortuna.output_calib_model.state import OutputCalibState
from fortuna.training.mixin import (
    InputValidatorMixin,
    WithCheckpointingMixin,
    WithEarlyStoppingMixin,
)
from fortuna.typing import (
    Array,
    Batch,
    CalibMutable,
    CalibParams,
    Path,
    Status,
)
from fortuna.utils.builtins import HashableMixin


class OutputCalibratorABC(
    HashableMixin,
    WithCheckpointingMixin,
    WithEarlyStoppingMixin,
    InputValidatorMixin,
    metaclass=abc.ABCMeta,
):
    def __init__(
        self,
        *args,
        calib_outputs_loader: TargetsLoader,
        predict_fn: Callable[[jnp.ndarray], jnp.ndarray],
        uncertainty_fn: Callable[[jnp.ndarray], jnp.ndarray],
        val_outputs_loader: Optional[TargetsLoader] = None,
        save_checkpoint_dir: Optional[Path] = None,
        save_every_n_steps: Optional[int] = None,
        keep_top_n_checkpoints: int = 2,
        disable_training_metrics_computation: bool = False,
        eval_every_n_epochs: int = 1,
        **kwargs,
    ):
        super(OutputCalibratorABC, self).__init__(*args, **kwargs)
        self._calib_outputs_loader = calib_outputs_loader
        self._val_outputs_loader = val_outputs_loader
        self.predict_fn = predict_fn
        self.uncertainty_fn = uncertainty_fn
        self.save_checkpoint_dir = save_checkpoint_dir
        self.save_every_n_steps = save_every_n_steps
        self.keep_top_n_checkpoints = keep_top_n_checkpoints
        self.disable_training_metrics_computation = disable_training_metrics_computation
        self.eval_every_n_epochs = eval_every_n_epochs
        self.multi_device = False

    def train(
        self,
        rng: PRNGKeyArray,
        state: OutputCalibState,
        loss_fun: Callable,
        training_data_loader: DataLoader,
        training_dataset_size: int,
        n_epochs: int = 1,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
        val_data_loader: Optional[DataLoader] = None,
        val_dataset_size: Optional[int] = None,
        verbose: bool = True,
    ) -> Tuple[OutputCalibState, Status]:
        training_losses_and_metrics = collections.defaultdict(list)
        val_losses_and_metrics = collections.defaultdict(list)

        state, data_loaders, outputs_loaders, rng = self.on_train_start(
            state,
            [training_data_loader, val_data_loader],
            [self._calib_outputs_loader, self._val_outputs_loader],
            rng,
        )
        training_data_loader, val_data_loader = data_loaders
        calib_outputs_loader, val_outputs_loader = outputs_loaders

        progress_bar = trange(n_epochs, desc="Epoch")
        for epoch in progress_bar:
            # training loop
            (
                state,
                training_losses_and_metrics_current_epoch,
                training_batch_metrics_str,
            ) = self._training_loop(
                epoch,
                loss_fun,
                metrics,
                rng,
                state,
                training_data_loader,
                calib_outputs_loader,
                training_dataset_size,
                verbose,
                progress_bar,
            )
            # keep track of training losses and metrics [granularity=epoch]
            for k in training_losses_and_metrics_current_epoch.keys():
                training_losses_and_metrics[k].append(
                    training_losses_and_metrics_current_epoch[k]
                )

            # validation loop
            if self.should_perform_validation(val_data_loader, epoch):
                # performance evaluation on the whole validation dataset
                state = self.on_val_start(state)
                (
                    val_losses_and_metrics_current_epoch,
                    val_epoch_metrics_str,
                ) = self._val_loop(
                    loss_fun=loss_fun,
                    metrics=metrics,
                    rng=rng,
                    state=state,
                    val_data_loader=val_data_loader,
                    val_outputs_loader=val_outputs_loader,
                    val_dataset_size=val_dataset_size,
                    verbose=verbose,
                )
                if verbose:
                    logging.info(f"Epoch: {epoch + 1} | " + val_epoch_metrics_str)
                # keep track of training losses and metrics [granularity=epoch] and check for early stopping
                for k in val_losses_and_metrics_current_epoch.keys():
                    val_losses_and_metrics[k].append(
                        val_losses_and_metrics_current_epoch[k]
                    )
                # check for early stopping
                if self.is_early_stopping_active and self._early_stopping.should_stop:
                    logging.info("[Early Stopping] Stopping training...")
                    break

        # aggregate
        training_status = {
            k: jnp.array(v) for k, v in training_losses_and_metrics.items()
        }
        val_status = {k: jnp.array(v) for k, v in val_losses_and_metrics.items()}
        status = dict(**training_status, **val_status)

        state = self.on_train_end(state)
        return state, status

    def _training_loop(
        self,
        current_epoch: int,
        loss_fun: Callable,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
        rng: PRNGKeyArray,
        state: OutputCalibState,
        training_data_loader: DataLoader,
        calib_outputs_loader: TargetsLoader,
        training_dataset_size: int,
        verbose: bool,
        progress_bar: TqdmDecorator,
    ) -> Tuple[OutputCalibState, Dict[str, float], str]:
        training_losses_and_metrics_epoch_all_steps = []
        training_batch_metrics_str = ""
        for step, (batch, outputs) in enumerate(
            zip(training_data_loader, calib_outputs_loader)
        ):
            # forward and backward pass
            state, aux = self.training_step(
                state, batch, outputs, loss_fun, rng, training_dataset_size
            )
            # compute training losses and metrics for the current batch
            training_losses_and_metrics_current_batch = self.training_step_end(
                current_epoch=current_epoch,
                state=state,
                aux=aux,
                batch=batch,
                metrics=metrics,
            )
            # keep track of training losses and metrics [granularity=batch]
            training_losses_and_metrics_epoch_all_steps.append(
                training_losses_and_metrics_current_batch
            )
            # logging
            if verbose:
                training_batch_metrics_str = " | ".join(
                    [
                        f"{m}: {round(float(v), 5)}"
                        for m, v in training_losses_and_metrics_current_batch.items()
                    ]
                )
                progress_bar.set_description(
                    f"Epoch: {current_epoch + 1} | " + training_batch_metrics_str,
                    refresh=True,
                )

        # compute training losses and metrics avg for the current epoch + other ops (if needed)
        training_losses_and_metrics_current_epoch = self.training_epoch_end(
            training_losses_and_metrics_epoch_all_steps
        )

        return (
            state,
            training_losses_and_metrics_current_epoch,
            training_batch_metrics_str,
        )

    def training_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[OutputCalibState, Dict[str, Any]]:
        # ensure to use a different key at each step
        model_key = random.fold_in(rng, state.step)

        grad_fn = value_and_grad(
            lambda params: self.training_loss_step(
                loss_fun, params, batch, outputs, state.mutable, model_key, n_data
            ),
            has_aux=True,
        )
        (loss, aux), grad = grad_fn(state.params)
        grad, loss = self.sync_gradients_and_loss(grad, loss)

        state = state.apply_gradients(grads=grad, mutable=aux["mutable"])
        return (
            state,
            {
                "loss": loss,
                "outputs": aux["outputs"],
                "logging_kwargs": aux["logging_kwargs"],
            },
        )

    @abc.abstractmethod
    def training_loss_step(
        self,
        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: CalibParams,
        batch: Batch,
        outputs: Array,
        mutable: CalibMutable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        pass

    def training_step_end(
        self,
        current_epoch: int,
        state: OutputCalibState,
        aux: Dict[str, Any],
        batch: Batch,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
    ) -> Dict[str, jnp.ndarray]:
        if (
            self.save_checkpoint_dir
            and self.save_every_n_steps
            and current_epoch % self.save_every_n_steps == 0
        ):
            self.save_checkpoint(
                state, self.save_checkpoint_dir, keep=self.keep_top_n_checkpoints
            )
        training_losses_and_metrics = {"loss": aux["loss"]}

        if aux["logging_kwargs"] is not None:
            for k, v in aux["logging_kwargs"].items():
                training_losses_and_metrics[k] = v

        if not self.disable_training_metrics_computation and metrics is not None:
            preds = self.predict_fn(aux["outputs"])
            uncertainties = self.uncertainty_fn(aux["outputs"])
            if self.multi_device:
                training_batch_metrics = self.compute_metrics(
                    preds.reshape((preds.shape[0] * preds.shape[1],) + preds.shape[2:]),
                    uncertainties.reshape(
                        (uncertainties.shape[0] * uncertainties.shape[1],)
                        + uncertainties.shape[2:]
                    ),
                    batch[1].reshape(
                        (batch[1].shape[0] * batch[1].shape[1],) + batch[1].shape[2:]
                    ),
                    metrics,
                )
            else:
                training_batch_metrics = self.compute_metrics(
                    preds, uncertainties, batch[1], metrics
                )
            for k, v in training_batch_metrics.items():
                training_losses_and_metrics[k] = v
        return training_losses_and_metrics

    def _val_loop(
        self,
        loss_fun: Callable,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
        rng: PRNGKeyArray,
        state: OutputCalibState,
        val_data_loader: DataLoader,
        val_outputs_loader: TargetsLoader,
        val_dataset_size: int,
        verbose: bool = True,
    ) -> Tuple[Dict[str, float], str]:
        val_losses_and_metrics_epoch_all_steps = []
        val_epoch_metrics_str = ""
        for batch, outputs in zip(val_data_loader, val_outputs_loader):
            val_losses_and_metrics_current_batch = self.val_step(
                state,
                batch,
                outputs,
                loss_fun,
                rng,
                val_dataset_size,
                metrics,
            )
            val_losses_and_metrics_epoch_all_steps.append(
                val_losses_and_metrics_current_batch
            )
        # compute validation losses and metrics for the current epoch
        val_losses_and_metrics_current_epoch = self.val_epoch_end(
            val_losses_and_metrics_epoch_all_steps, state
        )
        # logging
        if verbose:
            val_epoch_metrics_str = " | ".join(
                [
                    f"{m}: {round(float(v), 5)}"
                    for m, v in val_losses_and_metrics_current_epoch.items()
                ]
            )
        return val_losses_and_metrics_current_epoch, val_epoch_metrics_str

    def val_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
    ) -> Dict[str, jnp.ndarray]:
        val_loss, aux = self.val_loss_step(state, batch, outputs, loss_fun, rng, n_data)
        val_metrics = self.val_metrics_step(aux, batch, metrics)
        return {"val_loss": val_loss, **val_metrics}

    @abc.abstractmethod
    def val_loss_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]:
        pass

    def val_metrics_step(
        self,
        aux: Dict[str, jnp.ndarray],
        batch: Batch,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
    ) -> Dict[str, jnp.ndarray]:
        if metrics is not None:
            val_metrics = self.compute_metrics(
                self.predict_fn(aux["outputs"]),
                self.uncertainty_fn(aux["outputs"]),
                batch[1],
                metrics,
            )
            return {f"val_{m}": v for m, v in val_metrics.items()}
        else:
            return {}

    def training_epoch_end(
        self, training_losses_and_metrics_current_epoch: List[Dict[str, jnp.ndarray]]
    ) -> Dict[str, float]:
        return self._get_mean_losses_and_metrics(
            training_losses_and_metrics_current_epoch
        )

    def val_epoch_end(
        self,
        val_losses_and_metrics_current_epoch: List[Dict[str, jnp.ndarray]],
        state: OutputCalibState,
    ) -> Dict[str, float]:
        val_losses_and_metrics_current_epoch = self._get_mean_losses_and_metrics(
            val_losses_and_metrics_current_epoch
        )
        # early stopping
        improved = self.early_stopping_update(val_losses_and_metrics_current_epoch)
        if improved and self.save_checkpoint_dir:
            self.save_checkpoint(state, self.save_checkpoint_dir, force_save=True)
        return val_losses_and_metrics_current_epoch

    def _get_mean_losses_and_metrics(
        self, losses_and_metrics: List[Dict[str, jnp.ndarray]]
    ) -> Dict[str, float]:
        losses_and_metrics = stack_forest(losses_and_metrics)
        losses_and_metrics = tree_map(lambda x: x.mean(), losses_and_metrics)
        return losses_and_metrics

    def should_perform_validation(
        self, val_data_loader: Optional[DataLoader], epoch: int
    ) -> bool:
        return (
            val_data_loader is not None
            and self.eval_every_n_epochs > 0
            and epoch % self.eval_every_n_epochs == 0
        )

    @staticmethod
    def sync_gradients_and_loss(
        grad: jnp.ndarray, loss: jnp.ndarray
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        return grad, loss

    def on_train_start(
        self,
        state: OutputCalibState,
        data_loaders: List[DataLoader],
        outputs_loaders: List[TargetsLoader],
        rng: PRNGKeyArray,
    ) -> Tuple[OutputCalibState, List[DataLoader], List[TargetsLoader], PRNGKeyArray]:
        return state, data_loaders, outputs_loaders, rng

    def on_train_end(self, state: OutputCalibState) -> OutputCalibState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )
        return state

    def on_val_start(self, state: OutputCalibState) -> OutputCalibState:
        return state

    def compute_metrics(
        self,
        preds: Array,
        uncertainties: Array,
        targets: Array,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
    ) -> Dict[str, Array]:
        metrics_vals = {}
        for metric in metrics:
            metrics_vals[metric.__name__] = metric(preds, uncertainties, targets)
        return metrics_vals


class JittedMixin:
    @partial(jax.jit, static_argnums=(0, 4, 6))
    def training_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[OutputCalibState, Dict[str, Any]]:
        return super().training_step(state, batch, outputs, loss_fun, rng, n_data)

    @partial(jax.jit, static_argnums=(0, 4, 6))
    def val_loss_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Dict[str, jnp.ndarray]:
        return super().val_loss_step(state, batch, outputs, loss_fun, rng, n_data)


class MultiDeviceMixin:
    all_reduce_mean = jax.pmap(lambda x: lax.pmean(x, "x"), "x")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.multi_device = True

    @staticmethod
    def _add_device_dim_to_data_loader(data_loader: DataLoader) -> DataLoader:
        def _reshape_batch(batch):
            n_devices = jax.local_device_count()
            if batch.shape[0] % n_devices != 0:
                raise ValueError(
                    f"The size of all batches must be a multiple of {n_devices}, that is the number of "
                    f"available devices. However, a batch with shape {batch.shape[0]} was found. "
                    f"Please set an appropriate batch size."
                )
            return batch.reshape((n_devices, -1) + batch.shape[1:])

        class DataLoaderWrapper:
            def __init__(self, data_loader: DataLoader):
                self._data_loader = data_loader

            def __iter__(self):
                data_loader = map(
                    lambda batch: tree_map(_reshape_batch, batch), self._data_loader
                )
                data_loader = jax_utils.prefetch_to_device(data_loader, 2)
                yield from data_loader

        return (
            DataLoaderWrapper(data_loader) if data_loader is not None else data_loader
        )

    @staticmethod
    def _add_device_dim_to_outputs_loader(
        outputs_loader: TargetsLoader,
    ) -> TargetsLoader:
        def _reshape_batch(batch):
            n_devices = jax.local_device_count()
            if batch.shape[0] % n_devices != 0:
                raise ValueError(
                    f"The size of all output batches must be a multiple of {n_devices}, that is the number of "
                    f"available devices. However, a batch of outputs with shape {batch.shape[0]} was found. "
                    f"Please set an appropriate batch size."
                )
            return batch.reshape((n_devices, -1) + batch.shape[1:])

        class TargetsLoaderWrapper:
            def __init__(self, outputs_loader: TargetsLoader):
                self._outputs_loader = outputs_loader

            def __iter__(self):
                outputs_loader = map(
                    lambda batch: tree_map(_reshape_batch, batch), self._outputs_loader
                )
                outputs_loader = jax_utils.prefetch_to_device(outputs_loader, 2)
                yield from outputs_loader

        return (
            TargetsLoaderWrapper(outputs_loader)
            if outputs_loader is not None
            else outputs_loader
        )

    @staticmethod
    def sync_mutable(state: OutputCalibState) -> OutputCalibState:
        return (
            state.replace(mutable=MultiDeviceMixin.all_reduce_mean(state.mutable))
            if state.mutable["output_calibrator"] is not None
            else state
        )

    @staticmethod
    def sync_gradients_and_loss(
        grads: jnp.ndarray, loss: jnp.ndarray
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        grad = lax.pmean(grads, axis_name="batch")
        loss = lax.pmean(loss, axis_name="batch")
        return grad, loss

    def save_checkpoint(
        self,
        state: OutputCalibState,
        save_checkpoint_dir: Path,
        keep: int = 1,
        force_save: bool = False,
        prefix: str = "checkpoint_",
    ) -> None:
        state = self.sync_mutable(state)
        state = jax.device_get(tree_map(lambda x: x[0], state))
        return super(MultiDeviceMixin, self).save_checkpoint(
            state, save_checkpoint_dir, keep, force_save, prefix
        )

    def on_train_start(
        self,
        state: OutputCalibState,
        data_loaders: List[DataLoader],
        outputs_loaders: List[TargetsLoader],
        rng: PRNGKeyArray,
    ) -> Tuple[OutputCalibState, List[DataLoader], List[TargetsLoader], PRNGKeyArray]:
        state, data_loaders, outputs_loaders, rng = super(
            MultiDeviceMixin, self
        ).on_train_start(state, data_loaders, outputs_loaders, rng)
        state = jax_utils.replicate(state)
        data_loaders = [
            self._add_device_dim_to_data_loader(dl) if dl is not None else dl
            for dl in data_loaders
        ]
        outputs_loaders = [
            self._add_device_dim_to_outputs_loader(ol) if ol is not None else ol
            for ol in outputs_loaders
        ]
        model_key = random.split(rng, jax.local_device_count())
        return state, data_loaders, outputs_loaders, model_key

    def on_train_end(self, state: OutputCalibState) -> OutputCalibState:
        state = super(MultiDeviceMixin, self).on_train_end(state)
        return jax.device_get(tree_map(lambda x: x[0], state))

    @partial(jax.pmap, axis_name="batch", static_broadcasted_argnums=(0, 4, 6))
    def training_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[OutputCalibState, Dict[str, Any]]:
        return super().training_step(state, batch, outputs, loss_fun, rng, n_data)

    def training_step_end(
        self,
        current_epoch: int,
        state: OutputCalibState,
        aux: Dict[str, Any],
        batch: Batch,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]],
    ) -> Dict[str, jnp.ndarray]:
        training_losses_and_metrics = super(MultiDeviceMixin, self).training_step_end(
            current_epoch, state, aux, batch, metrics
        )
        return tree_map(lambda x: x.mean(), training_losses_and_metrics)

    def on_val_start(self, state: OutputCalibState) -> OutputCalibState:
        state = super(MultiDeviceMixin, self).on_val_start(state)
        if state.mutable["output_calibrator"] is not None:
            state = self.sync_mutable(state)
        return state

    @partial(jax.pmap, axis_name="batch", static_broadcasted_argnums=(0, 4, 6))
    def val_loss_step(
        self,
        state: OutputCalibState,
        batch: Batch,
        outputs: Array,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Dict[str, jnp.ndarray]:
        val_losses = super().val_loss_step(state, batch, outputs, loss_fun, rng, n_data)
        return lax.pmean(val_losses, axis_name="batch")

    def val_metrics_step(
        self,
        aux: Dict[str, jnp.ndarray],
        batch: Batch,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
    ) -> Dict[str, jnp.ndarray]:
        outputs = aux["outputs"]
        outputs = outputs.reshape(outputs.shape[0] * outputs.shape[1], -1)
        targets = batch[1].reshape(batch[1].shape[0] * batch[1].shape[1], -1)
        if metrics is not None:
            val_metrics = self.compute_metrics(
                self.predict_fn(outputs), self.uncertainty_fn(outputs), targets, metrics
            )
            return {f"val_{m}": v for m, v in val_metrics.items()}
        else:
            return {}



import torch
from torch.nn import CrossEntropyLoss


@torch.no_grad()
def perplexity(logits: torch.Tensor, labels: torch.Tensor, n_final_tokens: int):
    loss_fct = CrossEntropyLoss(reduction="none")
    shift_logits = logits[..., :-1, :].contiguous()
    shift_labels = labels[..., 1:].contiguous()

    return torch.exp(
        loss_fct(shift_logits.transpose(1, 2), shift_labels)[:, -n_final_tokens:].mean(
            1
        )
    )


@torch.no_grad()
def inv_perplexity(logits: torch.Tensor, labels: torch.Tensor, n_final_tokens: int):
    return 1 / perplexity(logits=logits, labels=labels, n_final_tokens=n_final_tokens)



from typing import Optional

from fortuna.typing import Path


class CalibCheckpointer:
    def __init__(
        self,
        save_checkpoint_dir: Optional[Path] = None,
        restore_checkpoint_path: Optional[Path] = None,
        save_every_n_steps: Optional[int] = None,
        keep_top_n_checkpoints: Optional[int] = 2,
        dump_state: bool = False,
    ):
        """
        An object to configure saving and restoring of checkpoints during the calibration process.

        Parameters
        ----------
        save_checkpoint_dir: Optional[Path] = None
            Save directory location.
        restore_checkpoint_path: Optional[Path]
            Path to checkpoint file or directory to restore.
        save_every_n_steps: int
            Number of training steps between checkpoints. To disable, set `every_n_train_steps` to None or 0 (no
            checkpoint will be saved during training).
        keep_top_n_checkpoints: int
            Number of past checkpoint files to keep.
        dump_state: bool
            Dump the fitted calibration state as a checkpoint in `save_checkpoint_dir`.
            Any future call to the state will internally involve restoring it from memory.
        """
        self.save_checkpoint_dir = save_checkpoint_dir
        self.save_every_n_steps = save_every_n_steps
        self.restore_checkpoint_path = restore_checkpoint_path
        self.keep_top_n_checkpoints = keep_top_n_checkpoints
        self.dump_state = dump_state



import jax.numpy as jnp

from fortuna.calib_model.predictive.base import Predictive
from fortuna.data.loader import InputsLoader
from fortuna.likelihood.classification import ClassificationLikelihood


class ClassificationPredictive(Predictive):
    def __init__(self, likelihood: ClassificationLikelihood):
        super().__init__(likelihood=likelihood)

    def entropy(
        self,
        inputs_loader: InputsLoader,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive entropy for each input.
        """
        state = self.state.get()
        return self.likelihood.entropy(
            params=state.params,
            inputs_loader=inputs_loader,
            mutable=state.mutable,
            distribute=distribute,
        )



from __future__ import annotations

from typing import (
    Dict,
    List,
    Optional,
)

import jax.numpy as jnp

from fortuna.prob_model.posterior.normalizing_flow.normalizing_flow_state import (
    NormalizingFlowState,
)
from fortuna.typing import Array
from fortuna.utils.strings import convert_string_to_tuple


class ADVIState(NormalizingFlowState):
    """
    Attributes
    ----------
    encoded_name: jnp.ndarray
        ADVI state name encoded as an array.
    """

    encoded_name: tuple = convert_string_to_tuple("ADVIState")
    _encoded_which_params: Optional[Dict[str, List[Array]]] = None



import jax.numpy as jnp

from fortuna.typing import Array


def root_mean_squared_error(preds: Array, targets: Array) -> float:
    """
    Compute the root-mean-squared error (RMSE).

    Parameters
    ----------
    preds: Array
        A two-dimensional array of predictions over the data points.
    targets: Array
        A two-dimensional array of target variables.

    Returns
    -------
    float
        The computed RMSE.
    """
    return jnp.sqrt(jnp.mean(jnp.sum((preds - targets) ** 2, axis=1)))


def rmse(preds: Array, targets: Array) -> float:
    """See :func:`.root_mean_squared_error`."""
    return root_mean_squared_error(preds, targets)


def mean_squared_error(preds: Array, targets: Array) -> Array:
    """
    Compute the mean-squared error (MSE).

    Parameters
    ----------
    preds: Array
        A two-dimensional array of predictions over the data points.
    targets: Array
        A two-dimensional array of target variables.

    Returns
    -------
    Array
        The computed MSE.
    """
    return jnp.mean(jnp.sum((preds - targets) ** 2, axis=1))


def mse(preds: Array, targets: Array) -> float:
    """See :func:`.mean_squared_error`."""
    return mean_squared_error(preds, targets)


def root_mean_absolute_error(preds: Array, targets: Array) -> float:
    """
    Compute the root-mean-absolute error (RMAE).

    Parameters
    ----------
    preds: Array
        A two-dimensional array of predictions over the data points.
    targets: Array
        A two-dimensional array of target variables.

    Returns
    -------
    float
        The computed RMAE.
    """
    return jnp.sqrt(jnp.mean(jnp.sum(jnp.abs(preds - targets), axis=1)))


def rmae(preds: Array, targets: Array) -> float:
    """See :func:`.root_mean_absolute_error`."""
    return root_mean_absolute_error(preds, targets)


def mean_absolute_error(preds: Array, targets: Array) -> Array:
    """
    Compute the mean-absolute error (MAE).

    Parameters
    ----------
    preds: Array
        A two-dimensional array of predictions over the data points.
    targets: Array
        A two-dimensional array of target variables.

    Returns
    -------
    Array
        The computed MAE.
    """
    return jnp.mean(jnp.sum(jnp.abs(preds - targets), axis=1))


def mae(preds: Array, targets: Array) -> float:
    """See :func:`.mean_absolute_error`."""
    return mean_absolute_error(preds, targets)


def prediction_interval_coverage_probability(
    lower_bounds: Array, upper_bounds: Array, targets: Array
) -> Array:
    """
    Compute the prediction interval coverage probability (PICP). This is the fraction of data points for which the
    true targets lie within the estimated interval. This is supported only for scalar target data.

    Parameters
    ----------
    lower_bounds: Array
        Predictive lower bounds. These are the lower bounds of the estimated predictive intervals. This can either
        be a one-dimensional array with entry corresponding to different data points, or a two-dimensional array
        with first axis corresponding to different data points, and second axis with only one dimension.
    upper_bounds: Array
        Predictive upper bounds. These are the upper bounds of the estimated predictive intervals. This can either
        be a one-dimensional array with entry corresponding to different data points, or a two-dimensional array
        with first axis corresponding to different data points, and second axis with only one dimension.
    targets: Array
        A two-dimensional array of target variables, or a one-dimensional array with second dimension of size 1.

    Returns
    -------
    Array
        The computed PICP.
    """
    if targets.ndim == 1:
        targets = targets[:, None]
    elif targets.shape[1] > 1:
        raise ValueError(
            """This metric is supported only for target data such that `target.shape[1] == 1`, but
        `target.shape[1] == {}` was found.""".format(
                targets.shape[1]
            )
        )
    if upper_bounds.ndim == 1:
        upper_bounds = upper_bounds[:, None]
    if upper_bounds.shape[1] != 1:
        raise ValueError(
            """The second axis of `upper_bounds` must contain only one component."""
        )
    if lower_bounds.ndim == 1:
        lower_bounds = lower_bounds[:, None]
    if lower_bounds.shape[1] != 1:
        raise ValueError(
            """The second axis of `lower_bounds` must contain only one component."""
        )

    satisfies_upper = targets <= upper_bounds
    satisfies_lower = targets >= lower_bounds
    return jnp.mean(satisfies_lower * satisfies_upper)


def picp(lower_bounds: Array, upper_bounds: Array, targets: Array) -> float:
    """See :func:`.prediction_interval_coverage_probability`."""
    return prediction_interval_coverage_probability(lower_bounds, upper_bounds, targets)



from typing import (
    Any,
    Callable,
    Optional,
    Tuple,
)

import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array

ModuleDef = Any


class MLP(nn.Module):
    """
    A multi-layer perceptron (MLP).

    Attributes
    ----------
    output_dim: int
        The output model dimension.
    widths: Tuple[int]
        The number of units of each hidden layer. Default: (30, 30).
    activations: Tuple[Callable[[Array], Array]]
        The activation functions after each hidden layer. Default: (flax.linen.relu, flax.linen.relu).
    dropout: ModuleDef
        Dropout module.
    dropout_rate: float
        Dropout rate.
    dense: ModuleDef
        Dense module.
    """

    output_dim: int
    widths: Optional[Tuple[int]] = (30, 30)
    activations: Optional[Tuple[Callable[[Array], Array]]] = (nn.relu, nn.relu)
    dropout: ModuleDef = nn.Dropout
    dropout_rate: float = 0.0
    dense: ModuleDef = nn.Dense

    def setup(self):
        if len(self.widths) != len(self.activations):
            raise Exception(
                "`widths` and `activations` must have the same number of elements."
            )
        self.dfe_subnet = MLPDeepFeatureExtractorSubNet(
            dense=self.dense,
            widths=self.widths,
            activations=self.activations[:-1],
            dropout=self.dropout,
            dropout_rate=self.dropout_rate,
        )
        self.output_subnet = MLPOutputSubNet(
            dense=self.dense,
            activation=self.activations[-1],
            output_dim=self.output_dim,
        )

    def __call__(self, x: Array, train: bool = False, **kwargs) -> jnp.ndarray:
        x = self.dfe_subnet(x, train)
        x = self.output_subnet(x)
        return x


class DeepResidualNet(MLP):
    """
    A multi-layer perceptron with residual connections
    """

    def setup(self):
        if len(self.widths) != len(self.activations):
            raise Exception(
                "`widths` and `activations` must have the same number of elements."
            )
        self.dfe_subnet = DeepResidualFeatureExtractorSubNet(
            dense=self.dense,
            widths=self.widths,
            activations=self.activations[:-1],
            dropout=self.dropout,
            dropout_rate=self.dropout_rate,
        )
        self.output_subnet = MLPOutputSubNet(
            dense=self.dense,
            activation=self.activations[-1],
            output_dim=self.output_dim,
        )


class MLPDeepFeatureExtractorSubNet(nn.Module):
    widths: Tuple[int]
    activations: Tuple[Callable[[Array], Array]]
    dense: ModuleDef = nn.Dense
    dropout: ModuleDef = nn.Dropout
    dropout_rate: float = 0.0

    """
    MLP Deep feature extractor sub-network.

    Attributes
    ----------
    widths: Tuple[int]
        The number of units of each hidden layer.
    activations: Tuple[Callable[[Array], Array]]
        The activation functions after each hidden layer.
    dense: ModuleDef
        Dense module.
    dropout: ModuleDef
        Dropout module.
    dropout_rate: float
        Dropout rate.
    """

    @nn.compact
    def __call__(self, x: Array, train: bool = False, **kwargs) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: Array
            Inputs.
        train: bool
            Whether it is training or inference.

        Returns
        -------
        jnp.ndarray
            Output of the hidden layers.
        """
        if hasattr(self, "spectral_norm"):
            dense = self.spectral_norm(self.dense, train=train)
        else:
            dense = self.dense
        dropout = self.dropout(self.dropout_rate)
        n_activations = len(self.activations)

        def update(i: int, x):
            x = dense(self.widths[i], name="hidden" + str(i + 1))(x)
            if i < n_activations:
                x = self.activations[i](x)
            x = dropout(x, deterministic=not train)
            return x

        x = x.reshape(x.shape[0], -1)
        for i in range(0, len(self.widths)):
            x = update(i, x)
        return x


class DeepResidualFeatureExtractorSubNet(MLPDeepFeatureExtractorSubNet):
    @nn.compact
    def __call__(self, x: Array, train: bool = False, **kwargs) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: Array
            Inputs.
        train: bool
            Whether it is training or prediction.

        Returns
        -------
        jnp.ndarray
            Output of the hidden layers.
        """
        if hasattr(self, "spectral_norm"):
            dense = self.spectral_norm(self.dense, train=train)
        else:
            dense = self.dense
        dropout = self.dropout(self.dropout_rate)
        n_activations = len(self.activations)

        def update(i: int, x):
            x = dense(self.widths[i], name="hidden" + str(i + 1))(x)
            if i < n_activations:
                x = self.activations[i](x)
            x = dropout(x, deterministic=not train)
            return x

        x = x.reshape(x.shape[0], -1)
        x = dense(self.widths[0], name="hidden" + str(0 + 1))(x)
        for i in range(1, len(self.widths)):
            h = jnp.copy(x)
            x = update(i, x)
            x = h + x
        return x


class MLPOutputSubNet(nn.Module):
    output_dim: int
    activation: Optional[Callable[[Array], Array]] = None
    dense: ModuleDef = nn.Dense
    """
    MLP output sub-network.

    Attributes
    ----------
    output_dim: int
        The output model dimension.
    activations: Tuple[Callable[[Array], Array]]
        The activation functions after each hidden layer.
    dense: ModuleDef
        Dense module.
    """

    @nn.compact
    def __call__(self, x: jnp.ndarray, **kwargs) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Outputs of the hidden layers.

        Returns
        -------
        jnp.ndarray
            Output of the last layer.
        """
        if self.activation is not None:
            x = self.activation(x)
        x = self.dense(self.output_dim, name="last")(x)
        return x



import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array


class LinearModel(nn.Module):
    """
    A linear model.

    Parameters
    ----------
    output_dim: int
        The output model dimension.
    """

    output_dim: int

    @nn.compact
    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        x = nn.Dense(self.output_dim, name="last")(x)
        return x



from __future__ import annotations

from typing import (
    Optional,
    Tuple,
)

import numpy as np

from fortuna.data.loader.base import (
    BaseDataLoaderABC,
    BaseInputsLoader,
    BaseTargetsLoader,
)
from fortuna.data.loader.utils import IterableData
from fortuna.typing import (
    Array,
    Batch,
    Shape,
)


class DataLoader(BaseDataLoaderABC):
    @property
    def num_unique_labels(self) -> Optional[int]:
        if self._num_unique_labels is None:
            self._num_unique_labels = len(np.unique(self.to_array_targets()))
        return self._num_unique_labels

    @classmethod
    def from_array_data(
        cls,
        data: Batch,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ) -> DataLoader:
        """
        Build a :class:`~fortuna.data.loader.DataLoader` object from a tuple of arrays of input and target variables,
        respectively.

        Parameters
        ----------
        data: Batch
            Input and target arrays of data.
        batch_size: Optional[int]
            The batch size. If not given, the data will not be batched.
        shuffle: bool
            Whether the data loader should shuffle at every call.
        prefetch: bool
            Whether to prefetch the next batch.

        Returns
        -------
        DataLoader
            A data loader built out of the tuple of arrays.
        """
        return cls(
            iterable=IterableData.from_batch_array_data(
                data, batch_size=batch_size, shuffle=shuffle, prefetch=prefetch
            )
        )

    def to_inputs_loader(self) -> InputsLoader:
        """
        Reduce a data loader to an inputs loader.

        Returns
        -------
        InputsLoader
            The inputs loader derived from the data loader.
        """
        return InputsLoader(IterableData.data_loader_to_inputs_iterable(self))

    def to_targets_loader(self) -> TargetsLoader:
        """
        Reduce a data loader to a targets loader.

        Returns
        -------
        TargetsLoader
            The targets loader derived from the data loader.
        """
        return TargetsLoader(IterableData.data_loader_to_targets_iterable(self))

    def to_array_data(self) -> Batch:
        """
        Reduce a data loader to a tuple of input and target arrays.

        Returns
        -------
        Batch
            Tuple of input and target arrays.
        """
        inputs, targets = [], []
        for batch_inputs, batch_targets in self:
            inputs.append(batch_inputs)
            targets.append(batch_targets)
        return np.concatenate(inputs, 0), np.concatenate(targets, 0)

    def to_array_inputs(self) -> Array:
        """
        Reduce a data loader to an array of target data.

        Returns
        -------
        Array
            Array of input data.
        """
        inputs = []
        for batch_inputs, batch_targets in self:
            inputs.append(batch_inputs)
        return np.concatenate(inputs, 0)

    def to_array_targets(self) -> Array:
        """
        Reduce a data loader to an array of target data.

        Returns
        -------
        Array
            Array of input data.
        """
        targets = []
        for batch_inputs, batch_targets in self:
            targets.append(batch_targets)
        return np.concatenate(targets, 0)

    def chop(self, divisor: int) -> DataLoader:
        """
        Chop the last part of each batch of the data loader, to make sure the number od data points per batch divides
        `divisor`.

        Parameters
        ----------
        divisor : int
            Number of data points that each batched must divide.

        Returns
        -------
        DataLoader
            A data loader with chopped batches.
        """

        def fun():
            for inputs, targets in self:
                reminder = targets.shape[0] % divisor
                if reminder == 0:
                    yield inputs, targets
                elif targets.shape[0] > divisor:
                    yield inputs[:-reminder], targets[:-reminder]

        return self.from_callable_iterable(fun)

    def split(self, n_data: int) -> Tuple[DataLoader, DataLoader]:
        """
        Split a data loader into two data loaders.

        Parameters
        ----------
        n_data: int
            Number of data point after which the data loader should be split. The first returned data loader will
            contain exactly `n_data` data points. The second one will contain the remaining ones.

        Returns
        -------
        Tuple[DataLoader, DataLoader]
            The two data loaders made out of the original one.
        """

        def data_loader1():
            count = 0
            for inputs, targets in self:
                if count == n_data:
                    break
                if count + inputs.shape[0] <= n_data:
                    count += inputs.shape[0]
                    yield inputs, targets
                else:
                    inputs, targets = (
                        inputs[: n_data - count],
                        targets[: n_data - count],
                    )
                    count = n_data
                    yield inputs, targets

        def data_loader2():
            count = 0
            for inputs, targets in self:
                if count > n_data:
                    yield inputs, targets
                elif (count <= n_data) and (count + inputs.shape[0] > n_data):
                    count2 = count
                    count += inputs.shape[0]
                    inputs, targets = (
                        inputs[n_data - count2 :],
                        targets[n_data - count2 :],
                    )
                    yield inputs, targets
                else:
                    count += inputs.shape[0]

        return self.from_callable_iterable(data_loader1), self.from_callable_iterable(
            data_loader2
        )

    def sample(self, seed: int, n_samples: int) -> DataLoader:
        """
        Sample from the data loader, with replacement.

        Parameters
        ----------
        seed: int
            Random seed.
        n_samples: int
            Number of samples.

        Returns
        -------
        DataLoader
            A data loader made of the sampled data points.
        """

        def fun():
            rng = np.random.default_rng(seed)
            count = 0

            while True:
                for inputs, targets in self:
                    if count == n_samples:
                        break
                    idx = rng.choice(2, inputs.shape[0]).astype("bool")
                    inputs, targets = inputs[idx], targets[idx]
                    if count + inputs.shape[0] > n_samples:
                        inputs, targets = (
                            inputs[: n_samples - count],
                            targets[: n_samples - count],
                        )
                    count += inputs.shape[0]
                    if inputs.shape[0] > 0:
                        yield inputs, targets

                if count == n_samples:
                    break

        return self.from_callable_iterable(fun)

    @property
    def input_shape(self) -> Shape:
        def fun():
            for inputs, targets in self:
                input_shape = inputs.shape[1:]
                break
            return input_shape

        return fun()


class InputsLoader(BaseInputsLoader):
    @classmethod
    def from_array_inputs(
        cls,
        inputs: Array,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ) -> InputsLoader:
        """
        Build a :class:`~fortuna.data.loader.InputsLoader` object from an array of input data.

        Parameters
        ----------
        inputs: Array
            Input array of data.
        batch_size: Optional[int]
            The batch size. If not given, the inputs will not be batched.
        shuffle: bool
            Whether the inputs loader should shuffle at every call.
        prefetch: bool
            Whether to prefetch the next batch.

        Returns
        -------
        InputsLoader
            An inputs loader built out of the array of inputs.
        """
        return cls(
            iterable=IterableData.from_array_data(
                inputs, batch_size=batch_size, shuffle=shuffle, prefetch=prefetch
            )
        )

    def to_array_inputs(self) -> Array:
        """
        Reduce an inputs loader to an array of inputs.

        Returns
        -------
        Array
            Array of input data.
        """
        inputs = []
        for batch_inputs in self:
            inputs.append(batch_inputs)
        return np.concatenate(inputs, 0)

    def chop(self, divisor: int) -> InputsLoader:
        """
        Chop the last part of each batch of the inputs loader, to make sure the number od data points per batch divides
        `divisor`.

        Parameters
        ----------
        divisor : int
            Number of data points that each batched must divide.

        Returns
        -------
        InputsLoader
            An inputs loader with chopped batches.
        """

        def fun():
            for inputs in self:
                reminder = inputs.shape[0] % divisor
                if reminder == 0:
                    yield inputs
                elif inputs.shape[0] > divisor:
                    yield inputs[:-reminder]

        return self.from_callable_iterable(fun)

    def sample(self, seed: int, n_samples: int) -> InputsLoader:
        """
        Sample from the inputs loader, with replacement.

        Parameters
        ----------
        seed: int
            Random seed.
        n_samples: int
            Number of samples.

        Returns
        -------
        InputsLoader
            An inputs loader made of the sampled inputs.
        """

        def fun():
            rng = np.random.default_rng(seed)
            count = 0

            while True:
                for inputs in self:
                    if count == n_samples:
                        break
                    idx = rng.choice(2, inputs.shape[0]).astype("bool")
                    inputs = inputs[idx]
                    if count + inputs.shape[0] > n_samples:
                        inputs = inputs[: n_samples - count]
                    count += inputs.shape[0]
                    if inputs.shape[0] > 0:
                        yield inputs

                if count == n_samples:
                    break

        return self.from_callable_iterable(fun)

    def split(self, n_data: int) -> Tuple[InputsLoader, InputsLoader]:
        """
        Split an inputs loader into two inputs loaders.

        Parameters
        ----------
        n_data: int
            Number of data point after which the inputs loader should be split. The first returned inputs loader will
            contain exactly `n_data` inputs. The second one will contain the remaining ones.

        Returns
        -------
        Tuple[InputsLoader, InputsLoader]
            The two inputs loaders made out of the original one.
        """

        def inputs_loader1():
            count = 0
            for inputs in self:
                if count == n_data:
                    break
                if count + inputs.shape[0] <= n_data:
                    count += inputs.shape[0]
                    yield inputs
                else:
                    inputs = inputs[: n_data - count]
                    count = n_data
                    yield inputs

        def inputs_loader2():
            count = 0
            for inputs in self:
                if count > n_data:
                    yield inputs
                elif (count <= n_data) and (count + inputs.shape[0] > n_data):
                    count2 = count
                    count += inputs.shape[0]
                    inputs = inputs[n_data - count2 :]
                    yield inputs
                else:
                    count += inputs.shape[0]

        return self.from_callable_iterable(inputs_loader1), self.from_callable_iterable(
            inputs_loader2
        )


class TargetsLoader(BaseTargetsLoader):
    @classmethod
    def from_array_targets(
        cls,
        targets: Array,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ) -> TargetsLoader:
        """
        Build a :class:`~fortuna.data.loader.TargetsLoader` object from an array of target data.

        Parameters
        ----------
        targets: Array
            Target array of data.
        batch_size: Optional[int]
            The batch size. If not given, the targets will not be batched.
        shuffle: bool
            Whether the target loader should shuffle at every call.
        prefetch: bool
            Whether to prefetch the next batch.

        Returns
        -------
        TargetsLoader
            A targets loader built out of the array of targets.
        """
        return cls(
            iterable=IterableData.from_array_data(
                targets, batch_size=batch_size, shuffle=shuffle, prefetch=prefetch
            )
        )

    def to_array_targets(self) -> Array:
        """
        Reduce a targets loader to an array of targets.

        Returns
        -------
        Array
            Array of target data.
        """
        targets = []
        for batch_targets in self:
            targets.append(batch_targets)
        return np.concatenate(targets, 0)

    def chop(self, divisor: int) -> TargetsLoader:
        """
        Chop the last part of each batch of the targets loader, to make sure the number od data points per batch divides
        `divisor`.

        Parameters
        ----------
        divisor : int
            Number of data points that each batched must divide.

        Returns
        -------
        TargetsLoader
            A targets loader with chopped batches.
        """

        def fun():
            for targets in self:
                reminder = targets.shape[0] % divisor
                if reminder == 0:
                    yield targets
                elif targets.shape[0] > divisor:
                    yield targets[:-reminder]

        return self.from_callable_iterable(fun)

    def sample(self, seed: int, n_samples: int) -> TargetsLoader:
        """
        Sample from the targets loader, with replacement.

        Parameters
        ----------
        seed: int
            Random seed.
        n_samples: int
            Number of samples.

        Returns
        -------
        TargetsLoader
            A targets loader made of the sampled targets.
        """

        def fun():
            rng = np.random.default_rng(seed)
            count = 0

            while True:
                for targets in self:
                    if count == n_samples:
                        break
                    idx = rng.choice(2, targets.shape[0]).astype("bool")
                    targets = targets[idx]
                    if count + targets.shape[0] > n_samples:
                        targets = targets[: n_samples - count]
                    count += targets.shape[0]
                    if targets.shape[0] > 0:
                        yield targets

                if count == n_samples:
                    break

        return self.from_callable_iterable(fun)

    def split(self, n_data: int) -> Tuple[TargetsLoader, TargetsLoader]:
        """
        Split a targets loader into two targets loaders.

        Parameters
        ----------
        n_data: int
            Number of data point after which the targets loader should be split. The first returned targets loader will
            contain exactly `n_data` targets. The second one will contain the remaining ones.

        Returns
        -------
        Tuple[TargetsLoader, TargetsLoader]
            The two targets loaders made out of the original one.
        """

        def targets_loader1():
            count = 0
            for targets in self:
                if count == n_data:
                    break
                if count + targets.shape[0] <= n_data:
                    count += targets.shape[0]
                    yield targets
                else:
                    targets = targets[: n_data - count]
                    count = n_data
                    yield targets

        def targets_loader2():
            count = 0
            for targets in self:
                if count > n_data:
                    yield targets
                elif (count <= n_data) and (count + targets.shape[0] > n_data):
                    count2 = count
                    count += targets.shape[0]
                    targets = targets[n_data - count2 :]
                    yield targets
                else:
                    count += targets.shape[0]

        return self.from_callable_iterable(
            targets_loader1
        ), self.from_callable_iterable(targets_loader2)
